import json
import os
from typing import Optional, Sequence

import requests  # type: ignore
from pydantic import BaseModel, ConfigDict

from .error import (
    APIError,
    AuthorizationError,
    RateLimitError,
    UnauthenticatedRateLimitError,
)

api_base_url_v1 = os.environ.get("ELEVEN_BASE_URL", "https://api.elevenlabs.io/v1")


class API(BaseModel):
    model_config = ConfigDict(
        # Parse enum to strings when converting to dict
        use_enum_values=True,
        # Allows `model_id` field
        protected_namespaces=(),
    )

    @staticmethod
    def request(url: str, method: str, api_key: Optional[str] = None, **kwargs):
        api_key = api_key or os.environ.get("ELEVEN_API_KEY")
        headers = {"xi-api-key": api_key}

        if method == "get":
            response = requests.get(url, headers=headers, **kwargs)
        elif method == "post":
            response = requests.post(url, headers=headers, **kwargs)
        elif method == "delete":
            response = requests.delete(url, headers=headers, **kwargs)
        else:
            raise ValueError(f"Invalid request method {method}")

        status_code = response.status_code

        if status_code == 200:
            return response

        # Parse the error message and status
        error = json.loads(response.text)
        message, status = "", ""
        if "detail" in error:
            detail = error["detail"]
            if isinstance(error, dict):
                message = detail.get("message", "")
                status = detail.get("status", "")
        else:
            message = str(error)
            status = str(response.status_code)

        # Raise the appropriate error
        if status_code == 401:
            if status == "quota_exceeded":
                if api_key is None:
                    raise UnauthenticatedRateLimitError(message)
                else:
                    raise RateLimitError(message)
            elif status == "needs_authorization":
                raise AuthorizationError(message)

        raise APIError(message, status)

    @staticmethod
    def get(url: str, *args, **kwargs):
        return API.request(url, method="get", *args, **kwargs)  # type: ignore

    @staticmethod
    def post(url: str, *args, **kwargs):
        return API.request(url, method="post", *args, **kwargs)  # type: ignore

    @staticmethod
    def delete(url: str, *args, **kwargs):
        return API.request(url, method="delete", *args, **kwargs)  # type: ignore


class Listable:
    @property
    def items(self) -> Sequence:
        raise NotImplementedError

    def __getitem__(self, idx: int):
        return self.items[idx]

    def __iter__(self):
        return iter(self.items)

    def __len__(self) -> int:
        return len(self.items)



from __future__ import annotations

import base64
import json
import os
from typing import Iterator, Optional

import websockets
from websockets.sync.client import connect

from .base import API, api_base_url_v1
from .model import Model
from .voice import Voice


def text_chunker(chunks: Iterator[str]) -> Iterator[str]:
    """Used during input streaming to chunk text blocks and set last char to space"""
    splitters = (".", ",", "?", "!", ";", ":", "—", "-", "(", ")", "[", "]", "}", " ")
    buffer = ""
    for text in chunks:
        if buffer.endswith(splitters):
            yield buffer if buffer.endswith(" ") else buffer + " "
            buffer = text
        elif text.startswith(splitters):
            output = buffer + text[0]
            yield output if output.endswith(" ") else output + " "
            buffer = text[1:]
        else:
            buffer += text
    if buffer != "":
        yield buffer + " "


class TTS(API):
    @staticmethod
    def generate(
        text: str, voice: Voice, model: Model, api_key: Optional[str] = None
    ) -> bytes:
        url = f"{api_base_url_v1}/text-to-speech/{voice.voice_id}"
        data = dict(
            text=text,
            model_id=model.model_id,
            voice_settings=voice.settings.model_dump() if voice.settings else None,
        )  # type: ignore
        response = API.post(url, json=data, api_key=api_key)
        return response.content

    @staticmethod
    def generate_stream(
        text: str,
        voice: Voice,
        model: Model,
        stream_chunk_size: int = 2048,
        api_key: Optional[str] = None,
        latency: int = 1,
    ) -> Iterator[bytes]:
        url = f"{api_base_url_v1}/text-to-speech/{voice.voice_id}/stream?optimize_streaming_latency={latency}"
        data = dict(
            text=text,
            model_id=model.model_id,
            voice_settings=voice.settings.model_dump() if voice.settings else None,
        )  # type: ignore
        response = API.post(url, json=data, stream=True, api_key=api_key)
        for chunk in response.iter_content(chunk_size=stream_chunk_size):
            if chunk:
                yield chunk

    @staticmethod
    def generate_stream_input(
        text: Iterator[str], voice: Voice, model: Model, api_key: Optional[str] = None
    ) -> Iterator[bytes]:
        BOS = json.dumps(
            dict(
                text=" ",
                try_trigger_generation=True,
                voice_settings=voice.settings.model_dump() if voice.settings else None,
                generation_config=dict(
                    chunk_length_schedule=[50],
                ),
            )
        )
        EOS = json.dumps(dict(text=""))

        with connect(
            f"wss://api.elevenlabs.io/v1/text-to-speech/{voice.voice_id}/stream-input?model_id={model.model_id}",
            additional_headers={
                "xi-api-key": api_key or os.environ.get("ELEVEN_API_KEY")
            },
        ) as websocket:
            # Send beginning of stream
            websocket.send(BOS)

            # Stream text chunks and receive audio
            for text_chunk in text_chunker(text):
                data = dict(text=text_chunk, try_trigger_generation=True)
                websocket.send(json.dumps(data))
                try:
                    data = json.loads(websocket.recv(1e-4))
                    if data["audio"]:
                        yield base64.b64decode(data["audio"])  # type: ignore
                except TimeoutError:
                    pass

            # Send end of stream
            websocket.send(EOS)

            # Receive remaining audio
            while True:
                try:
                    data = json.loads(websocket.recv())
                    if data["audio"]:
                        yield base64.b64decode(data["audio"])  # type: ignore
                except websockets.exceptions.ConnectionClosed:
                    break



from __future__ import annotations

import json
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import Field

from .base import API, Listable, api_base_url_v1
from .error import APIError


class UnauthorizedVoiceCloningError(APIError):
    pass


class VoiceSettings(API):
    stability: float = Field(..., ge=0.0, le=1.0)
    similarity_boost: float = Field(..., ge=0.0, le=1.0)
    style: Optional[float] = Field(None, ge=0.0, le=1.0)
    use_speaker_boost: Optional[bool] = None

    @classmethod
    def from_voice_id(cls, voice_id: str) -> VoiceSettings:
        url = f"{api_base_url_v1}/voices/{voice_id}/settings"
        return cls(**API.get(url).json())

    @classmethod
    def from_default(cls) -> VoiceSettings:
        url = f"{api_base_url_v1}/voices/settings/default"
        return cls(**API.get(url).json())


class VoiceSample(API):
    sample_id: str = ""
    file_name: str = ""
    mime_type: str = ""
    size_bytes: Optional[int] = None
    hash: str = ""


class VoiceClone(API):
    name: str = Field(..., min_length=1, max_length=100)
    description: str = ""
    files: List[str] = Field(..., min_length=1, max_length=25)
    labels: Optional[Dict[str, str]] = None

    @property
    def files_tuple(self):
        files_tuple = []
        for filepath in self.files:
            b = open(filepath, "rb")
            file_tuple = ("files", (f"{Path(filepath).stem}_{id(b)}", b, "audio/mpeg"))
            files_tuple.append(file_tuple)
        return files_tuple


class Gender(str, Enum):
    female = "female"
    male = "male"


class Age(str, Enum):
    young = "young"
    middle_aged = "middle_aged"
    old = "old"


class Accent(str, Enum):
    british = "british"
    american = "american"
    african = "african"
    australian = "australian"
    indian = "indian"


class VoiceDesign(API):
    name: str
    text: str = Field(..., min_length=100)
    voice_description: str
    gender: Gender
    age: Age
    accent: Accent
    accent_strength: float = Field(..., gt=0.3, lt=2.0)
    # The following fields are populated only after `generate` is called
    generated_voice_id: Optional[str] = None
    audio: Optional[bytes] = None

    def generate(self) -> bytes:
        url = f"{api_base_url_v1}/voice-generation/generate-voice"
        response = API.post(url, json=self.model_dump())
        self.generated_voice_id = response.headers["generated_voice_id"]
        self.audio = response.content
        return self.audio  # type: ignore


class Voice(API):
    voice_id: str
    name: Optional[str] = None
    category: Optional[str] = None
    description: Optional[str] = None
    labels: Optional[Dict[str, str]] = None
    samples: Optional[List[VoiceSample]] = None
    design: Optional[VoiceDesign] = None
    preview_url: Optional[str] = None
    settings: Optional[VoiceSettings] = None

    @classmethod
    def from_id(cls, voice_id: str):
        url = f"{api_base_url_v1}/voices/{voice_id}?with_settings=true"
        return cls(**API.get(url).json())

    @classmethod
    def from_clone(cls, voice_clone: VoiceClone) -> Voice:
        url = f"{api_base_url_v1}/voices/add"
        data = dict(
            name=voice_clone.name,
            description=voice_clone.description,
            labels=str(json.dumps(voice_clone.labels or {})),
        )
        files = voice_clone.files_tuple
        try:
            voice_id = API.post(url, data=data, files=files).json()["voice_id"]
        except APIError as e:
            if e.status == "can_not_use_instant_voice_cloning":
                raise UnauthorizedVoiceCloningError(e.message)
            raise
        return cls.from_id(voice_id)

    @classmethod
    def from_design(cls, voice_design: VoiceDesign):
        # If the voice design has not been generated yet, generate it
        if voice_design.generated_voice_id is None:
            voice_design.generate()
        # Create the voice from the voice design
        url = f"{api_base_url_v1}/voice-generation/create-voice"
        data = dict(
            voice_name=voice_design.name,
            generated_voice_id=voice_design.generated_voice_id,
            voice_description=voice_design.voice_description,
        )
        response = API.post(url, json=data)
        voice = cls.from_id(voice_id=response.json()["voice_id"])
        voice.design = voice_design
        return voice

    def fetch_settings(self):
        self.settings = VoiceSettings.from_voice_id(self.voice_id)

    @staticmethod
    def default_settings() -> VoiceSettings:
        return VoiceSettings.from_default()

    def delete(self):
        API.delete(f"{api_base_url_v1}/voices/{self.voice_id}")

    @classmethod
    def edit_by_id(
        cls,
        voice_id: str,
        name: Optional[str] = None,
        labels: Optional[Dict[str, str]] = None,
        description: Optional[str] = None,
        voice_settings: Optional[VoiceSettings] = None,
        voice_clone: Optional[VoiceClone] = None,
    ) -> Voice:
        url = f"{api_base_url_v1}/voices/{voice_id}/edit"
        data = {}
        data.update(dict(name=name) if name else {})
        data.update(dict(labels=str(labels)) if labels else {})
        data.update(dict(description=description) if description else {})

        files = None
        if voice_clone:
            clone_data = voice_clone.model_dump()
            clone_data["labels"] = str(clone_data.pop("labels"))
            del clone_data["files"]
            files = clone_data.pop("_files_tuple")

        if data or files:
            API.post(url, data=data, files=files)

        if voice_settings:
            settings_url = f"{api_base_url_v1}/voices/{voice_id}/settings/edit"
            API.post(settings_url, json=voice_settings.model_dump())

        return Voice.from_id(voice_id)

    def edit(self, **kwargs):
        updated_voice = Voice.edit_by_id(voice_id=self.voice_id, **kwargs)
        self.__dict__.update(updated_voice.model_dump())


class Voices(Listable, API):
    voices: List[Voice]

    @classmethod
    def from_api(cls, api_key: Optional[str] = None):
        url = f"{api_base_url_v1}/voices"
        response = API.get(url).json()
        return cls(**response)

    @property
    def items(self):
        return self.voices



import os
import re
from typing import Iterator, Optional, Union

from .api import TTS, Model, Voice, VoiceClone, Voices, VoiceSettings

DEFAULT_VOICE = Voice(
    voice_id="EXAVITQu4vr4xnSDxMaL",
    name="Bella",
    settings=VoiceSettings(
        stability=0.71, similarity_boost=0.5, style=0.0, use_speaker_boost=True
    ),
)


def set_api_key(api_key: str) -> None:
    os.environ["ELEVEN_API_KEY"] = api_key


def get_api_key() -> Optional[str]:
    return os.environ.get("ELEVEN_API_KEY")


def voices() -> Voices:
    return Voices.from_api()


def clone(**kwargs) -> Voice:
    return Voice.from_clone(VoiceClone(**kwargs))


def is_voice_id(val: str) -> bool:
    return bool(re.match(r"^[a-zA-Z0-9]{20}$", val))


def generate(
    text: Union[str, Iterator[str]],
    api_key: Optional[str] = None,
    voice: Union[str, Voice] = DEFAULT_VOICE,
    model: Union[str, Model] = "eleven_monolingual_v1",
    stream: bool = False,
    latency: int = 1,
    stream_chunk_size: int = 2048,
) -> Union[bytes, Iterator[bytes]]:

    if isinstance(voice, str):
        voice_str = voice
        # If voice is valid voice_id, use it
        if is_voice_id(voice):
            voice = Voice(voice_id=voice)
        else:
            voice = next((v for v in voices() if v.name == voice_str), None)  # type: ignore # noqa E501

        # Raise error if voice not found
        if not voice:
            raise ValueError(f"Voice '{voice_str}' not found.")

    if isinstance(model, str):
        model = Model(model_id=model)

    assert isinstance(voice, Voice)
    assert isinstance(model, Model)

    if stream:
        if isinstance(text, str):
            return TTS.generate_stream(
                text, voice, model, stream_chunk_size, api_key=api_key, latency=latency
            )  # noqa E501
        elif isinstance(text, Iterator):
            return TTS.generate_stream_input(text, voice, model, api_key=api_key)
    else:
        assert isinstance(text, str)
        return TTS.generate(text, voice, model, api_key=api_key)



from __future__ import annotations

from datetime import datetime
from typing import List, Optional

from pydantic import model_validator

from .base import API, Listable, api_base_url_v1
from .voice import VoiceSettings


class FeedbackItem(API):
    thumbs_up: bool
    feedback: str
    emotions: bool
    inaccurate_clone: bool
    glitches: bool
    audio_quality: bool
    other: bool
    review_status: str


class HistoryItem(API):
    history_item_id: str
    request_id: Optional[str] = None
    voice_id: str
    text: str
    date: Optional[datetime] = None
    date_unix: int
    character_count_change_from: int
    character_count_change_to: int
    character_count_change: Optional[int] = None
    content_type: str
    settings: Optional[VoiceSettings] = None
    feedback: Optional[FeedbackItem] = None
    _audio: Optional[bytes] = None

    @model_validator(mode="after")
    def add_computed_fields(self):
        # Compute character count field
        change_from = self.character_count_change_from
        change_to = self.character_count_change_to
        self.character_count_change = change_to - change_from
        # Compute datetime field
        self.date = datetime.utcfromtimestamp(self.date_unix)
        return self

    @classmethod
    def from_id(cls, history_item_id: str) -> HistoryItem:
        url = f"{api_base_url_v1}/history/{history_item_id}"
        response = API.get(url).json()
        return cls(**response)

    @property
    def audio(self) -> bytes:
        url = f"{api_base_url_v1}/history/{self.history_item_id}/audio"
        if self._audio is None:
            self._audio = API.get(url).content
        return self._audio

    def delete(self):
        API.delete(f"{api_base_url_v1}/history/{self.history_item_id}")


class History(Listable, API):
    last_history_item_id: Optional[str] = None
    history: List[HistoryItem]
    has_more: bool

    @classmethod
    def from_api(
        cls, page_size: int = 100, start_after_history_item_id: Optional[str] = None
    ) -> History:
        assert page_size < 1000, (
            "page_size must be less than 1000, change start_after_history_item_id"
            " instead."
        )
        data = dict(
            page_size=page_size,
            start_after_history_item_id=start_after_history_item_id,
        )
        url = f"{api_base_url_v1}/history"
        response = API.get(url, params=data).json()
        return cls(**response)

    @property
    def items(self):
        return self.history

    def __iter__(self):
        """Lazy iterator over history items"""
        for item in self.history:
            yield item
        while self.has_more:
            history_next = self.from_api(
                start_after_history_item_id=self.last_history_item_id
            )
            self.history.extend(history_next.history)
            self.has_more = history_next.has_more
            self.last_history_item_id = history_next.last_history_item_id

            for item in self.history:
                yield item



def test_model():
    from elevenlabs import Model, Models

    # Test that we can get all models
    models = Models.from_api()
    assert isinstance(models, Models)
    assert len(models) > 0
    assert isinstance(models[0], Model)



from ..utils import use_play


def test_voice_from_id():
    from elevenlabs import Voice, VoiceSettings

    # Test that we can get a voice from id
    voice_id = "21m00Tcm4TlvDq8ikWAM"
    voice = Voice.from_id(voice_id)
    assert isinstance(voice, Voice)

    assert voice.voice_id == voice_id
    assert voice.name == "Rachel"
    assert voice.category == "premade"
    assert isinstance(voice.settings, VoiceSettings)


def test_voice_clone():
    from elevenlabs import Voice, clone, generate, play

    from ..utils import as_local_files

    voice_file_urls = [
        "https://user-images.githubusercontent.com/12028621/235474694-584f7103-dab2-4c39-bb9a-8e5f00be85da.webm",
        "https://user-images.githubusercontent.com/12028621/235474694-584f7103-dab2-4c39-bb9a-8e5f00be85da.webm",
    ]

    with as_local_files(voice_file_urls) as files:
        voice = clone(
            name="Alex",
            description=(
                "An old American male voice with a slight hoarseness in his throat."
                " Perfect for news"
            ),
            files=files,
        )

    assert isinstance(voice, Voice)
    assert voice.voice_id is not None
    assert voice.name == "Alex"
    assert voice.category == "cloned"
    assert len(voice.samples) == len(voice_file_urls)

    audio = generate(
        text="Voice clone test successful.",
        voice=voice,
    )
    assert isinstance(audio, bytes) and len(audio) > 0
    voice.delete()

    if use_play:
        play(audio)


def test_voice_design():
    from elevenlabs import Accent, Age, Gender, Voice, VoiceDesign, generate, play

    voice_design = VoiceDesign(
        name="Lexa",
        text=(
            "Hi! My name is Lexa, I'm a voice design test. I should have a middle aged"
            " female voice with a british accent. "
        ),
        voice_description="Middle aged female with british accent.",
        gender=Gender.female,
        age=Age.middle_aged,
        accent=Accent.british,
        accent_strength=1.5,
    )

    audio = voice_design.generate()
    assert isinstance(audio, bytes) and len(audio) > 0
    if use_play:
        play(audio)

    voice = Voice.from_design(voice_design)
    assert isinstance(voice, Voice)

    audio = generate(
        text="Voice design test successful.",
        voice=voice,
    )
    assert isinstance(audio, bytes) and len(audio) > 0
    voice.delete()
    if use_play:
        play(audio)


def test_voices():
    from elevenlabs import Voice, Voices

    # Test that we can get voices from api
    voices = Voices.from_api()

    assert isinstance(voices, Voices)
    assert len(voices) > 0
    assert isinstance(voices[0], Voice)

    for voice in voices:
        assert isinstance(voice, Voice)



from typing import Optional


class APIError(Exception):
    message: str
    status: str
    display: Optional[str] = None

    def __init__(self, message: str, status: str = ""):
        self.message = message
        self.status = status
        self.display = self.display or message
        super().__init__(self.display)


class AuthorizationError(APIError):
    display: str = "This endpoint requires a valid API key, but none was found."


class RateLimitError(APIError):
    pass


class UnauthenticatedRateLimitError(RateLimitError):
    display: str = (
        "Thanks for trying out our speech synthesis! You have reached the limit of"
        " unauthenticated requests. You can continue, for free, by setting a valid API"
        " key."
    )
